# Copyright 2016-2018 Ioan Vancea, Markus Scheidgen
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flask import request, jsonify, make_response, abort
from . import api
from .. import db, http_auth
from ..models import *
from ..decorators import json, paginate
from ..utils import sum_form, check_value_type, formula_hill_as_string, formula_hill_as_list, reformat_user_formula
from sqlalchemy import func, text, or_
import re
from ..utils import validate_date, collect_insert_dbstats, generate_auth_token, verify_auth_token, flag_to_gitlab

# needed for RQ and RQScheduler
from redis import Redis
# from rq import Queue
# # for online server I need AUTH for redis
# # if 'enc-staging-nomad' in os.uname().nodename:
# #     q = Queue(connection=Redis(password='My_Pass'))
# # else:
# q = Queue(connection=Redis())

# needed for SAML authentication
# BEGIN here
import os
import socket
from flask import render_template, redirect, session, url_for
from urllib.parse import urlparse
from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.utils import OneLogin_Saml2_Utils

import app.config as config

# based on machine's hostname use different config files for SAML
# TODO switch to use environment variables: PRODUCTION, STAGING, DEVELOPMENT
hostname = socket.gethostname()
if hostname == 'enc-production-nomad':
    saml_config_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'saml-production')
elif hostname == 'enc-production-hv-nomad':
    saml_config_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'saml-production-hv')
elif hostname == 'enc-staging-nomad':
    saml_config_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'saml-staging')
else:
    saml_config_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'saml-testing')


def init_saml_auth(req):
    auth = OneLogin_Saml2_Auth(req, custom_base_path=saml_config_dir)
    return auth

def prepare_flask_request(request):
    # If server is behind proxys or balancers use the HTTP_X_FORWARDED fields
    url_data = urlparse(request.url)
    # On staging machine we still have HTTP by default so, I want the API app to
    # work when using SAML authentication with the IdP provider. That's why I need
    # different settings for this function
    if hostname == 'enc-production-nomad':
        return {
            'https': 'on' if request.scheme == 'https' else 'off',
            'http_host': request.host + '/api',
            'server_port': 443,
            'script_name': request.path,
            'get_data': request.args.copy(),
            # Uncomment if using ADFS as IdP, https://github.com/onelogin/python-saml/pull/144
            # 'lowercase_urlencoding': True,
            'post_data': request.form.copy()
        }
    elif hostname == 'enc-production-hv-nomad':
        return {
            'https': 'on' if request.scheme == 'https' else 'off',
            'http_host': request.host + '/api',
            'server_port': 443,
            'script_name': request.path,
            'get_data': request.args.copy(),
            # Uncomment if using ADFS as IdP, https://github.com/onelogin/python-saml/pull/144
            # 'lowercase_urlencoding': True,
            'post_data': request.form.copy()
        }
    else:
        return {
            'https': 'on' if request.scheme == 'https' else 'off',
            'http_host': request.host,
            'server_port': url_data.port,
            'script_name': request.path,
            'get_data': request.args.copy(),
            # Uncomment if using ADFS as IdP, https://github.com/onelogin/python-saml/pull/144
            # 'lowercase_urlencoding': True,
            'post_data': request.form.copy()
        }
# END here


# needed for HTTPAuth
# BEGIN here
@http_auth.verify_password
def verify_password(token, unused):
    # if token:
    user = verify_auth_token(token)
    if user:
        return True
    return False


@http_auth.error_handler
def unauthorized_token():
    response = jsonify({'status': 401, 'error': 'unauthorized',
                        'message': 'please send your authentication token',
                        'login_url': url_for('api.index', _external=True)})
    response.status_code = 401
    return response
# END here


############################
# Routes - URI
############################


@api.route('/')
@http_auth.login_required
def show_all():
    return make_response(jsonify({'error': 'Resource not found'}), 404)


# POST Method - Initial basic implementation
@api.route('/esmaterials', methods=['POST'])
# the authentication is disabled until the full search is completely implemented
# @http_auth.login_required
def get_es_materials():
    from elasticsearch import Elasticsearch
    es = Elasticsearch('http://localhost:9200')
    # get the content of the POSTed body
    json_data = request.get_json()
    if not json_data:
        return jsonify({'message': 'No input data provided'}), 400
    # checking the validity of the syntax of the elastic search POSTed query
    is_query_valid = es.indices.validate_query(index='aggregated', doc_type='search', body=json_data)
    if is_query_valid['valid'] is False:
        return jsonify(Error="Query syntax is not valid")
    # if the POSTed query syntax is valid we perform the query
    es_search_results = es.search(index='aggregated', doc_type='search', body=json_data)
    total_es_search_results = es_search_results['hits']['total']
    # print(total_es_search_results)
    # mat_id = str(es_search_results['hits']['hits'][0])
    materials_found_list = [hit['_source'] for hit in es_search_results['hits']['hits']]
    # print("=====here")
    # print(mat_id)
    # print("=====endhere")
    # print(ids)
    # return jsonify(es_search_results)
    # return jsonify(results=materials_found_list, total_results=total_es_search_results)
    myresults = es_search_material_schema.dump(materials_found_list).data
    return jsonify(results=myresults, total_results=total_es_search_results)


# Material
@api.route('/materials', methods=['POST'])
@http_auth.login_required
@json
@paginate('result', search_material_schema)
def get_materials():
    # get the content of the POSTed body
    json_data = request.get_json()
    if not json_data:
        return jsonify({'message': 'No input data provided'}), 400
    # collect the search criteria and exclusive flag
    search_criteria = json_data['search_by']
    # exclusive_search_flag = json_data['search_by']['exclusive']
    exclusive_search_flag = json_data['search_by'].get('exclusive', "1")

    """A dictionary to map the user submitted properties to their corresponding tables."""
    map_dict = {
        'material_name': Material.material_name,
        'structure_type': Material.structure_type,
        'space_group_number': Material.space_group_number,
        'system_type': Material.system_type,
        'crystal_system': Material.crystal_system,
        'band_gap': Calc.band_gap,
        'band_gap_direct': Calc.band_gap_direct,
        'has_band_structure': Calc.has_band_structure,
        'has_dos': Calc.has_dos,
        'has_fermi_surface': Calc.has_fermi_surface,
        'has_thermal_properties': Calc.has_thermal_properties,
        'functional_type': Calc.functional_type,
        'basis_set_type': Calc.basis_set_type,
        'mass_density': Calc.mass_density,
        'code_name': Calc.code_name
    }

    # I don't map the "search_by" key-value so there is
    # no need for the "search_by" entry in json anymore
    del json_data['search_by']

    # perform the mapping of properties to the Tables they belong in database
    mapped_dict = {map_dict[k]: v for k, v in json_data.items()}
    # print("===")
    # print(mapped_dict[Calc.code_name])
    # print(mapped_dict)

    # perform the search for each case
    if 'formula' in search_criteria:
        user_formula = search_criteria['formula']
        print("====", user_formula, "===")
        if exclusive_search_flag == "1":
            # sorted_user_formula = "".join(split_and_order(user_formula))
            # reformat the user's formula and use it as a string
            sorted_user_formula = "".join(reformat_user_formula(user_formula))
            # print(sorted_user_formula)
            result = Material.query.with_entities(Material.id, Material.formula, Material.formula_reduced, Material.structure_type, Material.space_group_number, Material.material_name, Material.system_type, func.count().label('nr_of_calculations_matching_criteria'), func.array_agg(Calc.id).label('calculations_list_matching_criteria')).\
                join(Calc).group_by(Material.id).having(Material.formula_reduced.like('{}'.format(sorted_user_formula))).\
                order_by(Material.formula_reduced, Material.structure_type).filter(*[check_value_type(attr, value) for attr, value in mapped_dict.items()])
            return result  # formula - exclusive search result
        else:
            # sorted_elements_in_formula = split_and_order(user_formula)
            # reformat the user's formula and use it as a list of strings
            sorted_elements_in_formula = reformat_user_formula(user_formula)
            result = Material.query.with_entities(Material.id, Material.formula, Material.formula_reduced, Material.structure_type, Material.space_group_number, Material.material_name, Material.system_type, func.count().label('nr_of_calculations_matching_criteria'), func.array_agg(Calc.id).label('calculations_list_matching_criteria')).\
                join(Calc).group_by(Material.id).filter(*[text("Material.formula_reduced ~ '{}([^0-9a-z]|$)'".format(X)) for X in sorted_elements_in_formula]).\
                order_by(Material.formula_reduced, Material.structure_type).filter(*[check_value_type(attr, value) for attr, value in mapped_dict.items()])
            return result  # formula - inclusive search result
    # dealing with element search
    elif 'element' in search_criteria:
        # user_element = request.args['element']
        user_element = search_criteria['element']
        # print(user_element)
        # print(type(user_element))
        # convert the elements to a LIST
        user_element_list = user_element.split(",")
        if exclusive_search_flag == "1":
            # we deal with exclusive search
            partial_search_string = ["{}\d*".format(x) for x in formula_hill_as_list(user_element_list)]
            full_search_string = "^" + ''.join(partial_search_string) + "$"
            result = Material.query.with_entities(Material.id, Material.formula, Material.formula_reduced, Material.structure_type, Material.space_group_number, Material.material_name, Material.system_type, func.count().label('nr_of_calculations_matching_criteria'), func.array_agg(Calc.id).label('calculations_list_matching_criteria')).\
                join(Calc).group_by(Material.id).\
                filter("Material.formula_reduced ~ '{}'".format(full_search_string)).\
                order_by(Material.formula_reduced, Material.structure_type).filter(*[check_value_type(attr, value) for attr, value in mapped_dict.items()])
            return result  # element - exclusive search result
        else:
            # we deal with inclusive search
            result = Material.query.with_entities(Material.id, Material.formula, Material.formula_reduced, Material.structure_type, Material.space_group_number, Material.material_name, Material.system_type, func.count().label('nr_of_calculations_matching_criteria'), func.array_agg(Calc.id).label('calculations_list_matching_criteria')).\
                join(Calc).group_by(Material.id).\
                filter(*[text("Material.formula_reduced ~ '{}([^a-z]|$)'".format(X)) for X in user_element_list]).\
                order_by(Material.formula_reduced, Material.structure_type).filter(*[check_value_type(attr, value) for attr, value in mapped_dict.items()])
            return result  # element inclusive search result
    # the case when the user doesn't introduce any element or formula
    else:
        result = Material.query.with_entities(Material.id, Material.formula, Material.formula_reduced, Material.structure_type, Material.space_group_number, Material.material_name, Material.system_type, func.count().label('nr_of_calculations_matching_criteria'), func.array_agg(Calc.id).label('calculations_list_matching_criteria')).\
                join(Calc).group_by(Material.id).order_by(Material.formula_reduced, Material.structure_type).\
                filter(*[check_value_type(attr, value) for attr, value in mapped_dict.items()])
        return result  # search result when no element or formula given
        # abort(404)  # return the error handler


@api.route('/materials/<int:material_id>', methods=['GET'])
@http_auth.login_required
# @json
def get_material(material_id):
    if 'property' in request.args:
        columns_list = [column.key for column in Material.__table__.columns]
        # print(columns_list)
        searched_column = request.args['property']
        if searched_column in columns_list:
            result = Material.query.with_entities(getattr(Material, searched_column)).filter(Material.id == material_id).first()
            return jsonify(material_schema.dump(result).data)
        else:
            return jsonify({'message': 'Requested property is not available'}), 400
    else:
        # to be used with @json decorator
        # return Material.query.get_or_404(material_id)
        results = Material.query.filter(Material.id == material_id).first()
        results = material_schema.dump(results).data
        return jsonify(results)


# TODO Modify the queries in the next route to not have duplicates
@api.route('/materials/<int:material_id>/groups', methods=['GET'])
@http_auth.login_required
def get_calcs_groups(material_id):
    query_eos = """
        select *
        from (
            select method_hash,
                material_hash,
                group_eos_hash,
                count(*) nr_of_calculations,
                array_agg(calc.id order by calc.cell_volume) as calculations_list
            from __schema__.material
            join __schema__.calc on material.id = calc.material_id
            join __schema__.energy on calc.id = energy.calc_id
            where material.id = %s
            and energy.e_kind='Total E'
            group by material_hash,
                  method_hash, group_eos_hash
            having count(*) > 4
        ) d
        cross join lateral (
            select energy.calc_id representative_calculation_id,
                energy.e_val energy_minimum
            from __schema__.energy
            where energy.calc_id = any(d.calculations_list)
              and energy.e_kind='Total E'
            order by energy.e_val
            limit 1
        ) last_energy;
    """.replace('__schema__', config.schema)
    query_convergence = """
        select *
        from (
            select method_hash,
                material_hash,
                group_parametervariation_hash,
                count(*) nr_of_calculations,
                array_agg(calc.id order by energy.e_val) as calculations_list
            from __schema__.material
            join __schema__.calc on material.id = calc.material_id
            join __schema__.energy on calc.id = energy.calc_id
            where material.id = %s
            and energy.e_kind='Total E'
            group by material_hash,
                  method_hash, group_parametervariation_hash
            having count(*) > 2
        ) d
        cross join lateral (
            select energy.calc_id representative_calculation_id,
                energy.e_val energy_minimum
            from __schema__.energy
            where energy.calc_id = any(d.calculations_list)
              and energy.e_kind='Total E'
            order by energy.e_val
            limit 1
        ) last_energy;
    """.replace('__schema__', config.schema)
    # we renamed group_convergencetest_hash_hash to group_parametervariation_hash
    # so, perhaps some var renaming should be done here
    mat_id = "{}".format(material_id)
    result_eos = db.engine.execute(query_eos, (mat_id))
    result_convergence = db.engine.execute(query_convergence, (mat_id))
    # result_method = db.engine.execute(query_method, (mat_id))
    # counting the number of groups found for each group type:
    # "rowcount" is a property not a method. This property could
    # be DBAPI specific, has to be checked
    # total_groups = result_eos.rowcount + result_convergence.rowcount + result_method.rowcount
    total_groups = result_eos.rowcount + result_convergence.rowcount
    # result_method_dump = search_groups_schema.dump(result_method).data
    result_eos_dump = search_groups_schema.dump(result_eos).data
    result_convergence_dump = search_groups_schema.dump(result_convergence).data
    # result = result_eos_dump + result_convergence_dump + result_method_dump
    result = result_eos_dump + result_convergence_dump
    return jsonify(total_groups=total_groups, groups=result)
    # return jsonify(total_groups=total_groups, group_method=result_method_dump,
    #                group_eos=result_eos_dump, group_convergence=result_convergence_dump)


# Element/Position
@api.route('/materials/<int:material_id>/elements', methods=['GET'])
@http_auth.login_required
@json
@paginate('element', m_element_schema)
def get_elements(material_id):
    material = Material.query.get_or_404(material_id)
    return material.element


@api.route('/materials/<int:material_id>/elements/<int:element_id>', methods=['GET'])
@http_auth.login_required
@json
def get_element(material_id, element_id):
    material = Material.query.get_or_404(material_id)
    element = material.element.filter(Element.id == element_id).first()
    return element


# Cell
@api.route('/materials/<int:material_id>/cells', methods=['GET'])
@http_auth.login_required
@json
@paginate('cell', m_cell_schema)
def get_cells(material_id):
    material = Material.query.get_or_404(material_id)
    return material.cell


@api.route('/materials/<int:material_id>/cells/<int:cell_id>', methods=['GET'])
@http_auth.login_required
@json
def get_cell(material_id, cell_id):
    material = Material.query.get_or_404(material_id)
    cell = material.cell.filter(Cell.id == cell_id).first()
    return cell


# Calc
@api.route('/materials/<int:material_id>/calculations', methods=['GET'])
@http_auth.login_required
@json
@paginate('calc', m_calc_schema)
def get_calcs(material_id):
    # material = Material.query.get_or_404(material_id)
    # return material.calc
    result = Calc.query.filter_by(material_id=material_id)
    return result


@api.route('/materials/<int:material_id>/calculations/<int:calc_id>', methods=['GET'])
@http_auth.login_required
# @json
def get_calc(material_id, calc_id):
    if 'property' in request.args:
        columns_list = [column.key for column in Calc.__table__.columns]
        json_values_list = DiagramData.data_type.property.columns[0].type.enums
        # print(columns_list)
        # print(json_values_list)
        searched_column = request.args['property']
        if searched_column in json_values_list:
            # print(searched_column)
            # result = DiagramData.query.with_entities(DiagramData.data_content)\
            #                           .filter(DiagramData.calc_id == calc_id)\
            #                           .filter(DiagramData.data_type == searched_column)\
            #                           .first()
            # need to join the Calc table to match the material_id, as well
            # a label is added dynamically for "data_content" to match the property
            # searched for, which is useful for Marshmallow model
            result = DiagramData.query.join(Calc).with_entities(DiagramData.data_content.label(searched_column))\
                                .filter(Calc.id == calc_id, Calc.material_id == material_id)\
                                .filter(DiagramData.calc_id == calc_id)\
                                .filter(DiagramData.data_type == searched_column)\
                                .first()
            return jsonify(calc_schema.dump(result).data)
        elif searched_column in columns_list:
            result = Calc.query.with_entities(getattr(Calc, searched_column)).filter(
                Calc.id == calc_id, Calc.material_id == material_id).first()
            return jsonify(calc_schema.dump(result).data)
        else:
            return jsonify({'message': 'Requested property is not available'}), 400
    else:
        # material = Material.query.get_or_404(material_id)
        # calculation = material.calc.filter(Calc.id==calc_id).first()
        calculation = Calc.query.filter(
            Calc.id == calc_id, Calc.material_id == material_id).first()
        # # next block is needed only if there are results from above query
        # # helper to flatten the nested diagram_data structure in the response JSON
        # if calculation is not None:
        #     for row in calculation.diagram_data:
        #         row_key = row.data_type
        #         row_content = row.data_content
        #         setattr(calculation, row_key, row_content)
        results = calc_schema.dump(calculation).data
        return jsonify(results)


# Energy_decomposition
@api.route('/materials/<int:material_id>/calculations/<int:calc_id>/energies', methods=['GET'])
@http_auth.login_required
@json
@paginate('results', m_energy_decomposition_schema)
def get_energies(material_id, calc_id):
    results = EnergyDecomposition.query.filter(EnergyDecomposition.calc.has(
        (Calc.id == calc_id) & (Calc.material_id == material_id)))
    return results


@api.route('/materials/<int:material_id>/calculations/<int:calc_id>/energies/<int:energy_id>', methods=['GET'])
@http_auth.login_required
@json
def get_energy(material_id, calc_id, energy_id):
    results = EnergyDecomposition.query.filter(EnergyDecomposition.id == energy_id, EnergyDecomposition.calc.has(
        (Calc.id == calc_id) & (Calc.material_id == material_id))).first()
    return results


# # Author
# @api.route('/materials/<int:material_id>/calculations/<int:calc_id>/authors', methods=['GET'])
# @json
# @paginate('results', m_contributor_schema)
# def get_authors(material_id, calc_id):
#     results = Contributor.query.filter(Contributor.calc.has(
#         (Calc.id == calc_id) & (Calc.material_id == material_id)))
#     return results
#
#
# @api.route('/materials/<int:material_id>/calculations/<int:calc_id>/authors/<int:author_id>', methods=['GET'])
# @json
# def get_author(material_id, calc_id, author_id):
#     results = Contributor.query.filter(Contributor.id == author_id, Contributor.calc.has(
#         (Calc.id == calc_id) & (Calc.material_id == material_id))).first()
#     return results


# Route to make data available for implementing GUI suggestions for some text fields
@api.route('/suggestions', methods=['GET'])
@http_auth.login_required
def get_suggestions():
    if request.args.get('property'):
        property = request.args['property']
        if property == "material_name":
            result = Material.query.with_entities(Material.material_name)\
                                   .distinct()\
                                   .filter(Material.material_name != None)\
                                   .order_by(Material.material_name)\
                                   .all()
            shownresult = [d[0] for d in result]
            return jsonify(material_name=shownresult)
        elif property == "structure_type":
            result = Material.query.with_entities(Material.structure_type)\
                                   .distinct()\
                                   .filter(Material.structure_type != "")\
                                   .order_by(Material.structure_type)\
                                   .all()
            shownresult = [d[0] for d in result]
            return jsonify(structure_type=shownresult)
        elif property == "space_group_number":
            result = Material.query.with_entities(Material.space_group_number)\
                                   .distinct()\
                                   .filter(Material.space_group_number != None)\
                                   .order_by(Material.space_group_number)\
                                   .all()
            shownresult = [d[0] for d in result]
            return jsonify(space_group_number=shownresult)
        elif property == 'code_name':
            result = Calc.query.with_entities(Calc.code_name)\
                               .distinct()\
                               .order_by(Calc.code_name)\
                               .all()
            shownresult = [d[0] for d in result]
            return jsonify(code_name=shownresult)
        # elif property == "contributors":
        #     result = Contributor.query.with_entities(Contributor.last_name, Contributor.first_name)\
        #                          .distinct()\
        #                          .order_by(Contributor.last_name)\
        #                          .all()
        #     shownresult = [', '.join(d) for d in result]
        #     return jsonify(contributors=shownresult)
        # elif property == "compound_class_springer":
        #     return make_response(jsonify({'error': 'compound_class_springer - not available yet'}), 404)
        # elif property == "material_class_springer":
        #     return make_response(jsonify({'error': 'material_class_springer - not available yet'}), 404)
        else:
            return make_response(jsonify({'error': 'Quantity not available'}), 404)
    else:
        abort(404)  # return the error handler


# Route to gather some useful statistics regarding the data available in
# encyclopedia database
@api.route('/dbstats', methods=['GET'])
@http_auth.login_required
def show_dbstats():
    if request.args.get('date') == 'first':
        # for one single entry
        stats = Stats.query.order_by(Stats.date.asc()).first()
        results = stats_schema.dump(stats).data
        return jsonify(results)
    if request.args.get('date') == 'last':
        # for one single entry
        stats = Stats.query.order_by(Stats.date.desc()).first()
        results = stats_schema.dump(stats).data
        return jsonify(results)
    if request.args.get('date') == 'all':
        # for multiple entries
        stats = Stats.query.order_by(Stats.date).all()
        results = m_stats_schema.dump(stats).data
        return jsonify(results)
    date_range = request.args.getlist('between', None)
    if len(date_range) == 2:
        # all_args = request.args.to_dict(flat=False)
        # return jsonify(all_args)
        # let's check if the date is formated as we want/need it
        if not validate_date(date_range[0]):
            return jsonify({'message': 'Start date is not correctly formatted: eg: 25-09-2017'}), 400
        if not validate_date(date_range[1]):
            return jsonify({'message': 'End date is not correctly formatted: eg: 25-09-2017'}), 400
        start_date = datetime.datetime.strptime(date_range[0], "%d-%m-%Y")
        end_date = datetime.datetime.strptime(date_range[1], "%d-%m-%Y")
        # check if the start and end date are given in correct order
        if start_date >= end_date:
            return jsonify({'message': 'Start date is after or identical with end date'}), 400
        # print(start_date, end_date)
        stats = Stats.query.filter(Stats.date.between(start_date, end_date)).order_by(Stats.date)
        results = m_stats_schema.dump(stats).data
        return jsonify(results)

    # Here to catch all other strange query parameters
    response = jsonify({'status': 400, 'error': 'bad request',
                        'message': 'Bad request'})
    response.status_code = 400
    return response


@api.route('/dbstats-collect', methods=['GET'])
@http_auth.login_required
def collect_dbstats():
    # get_stats_results = q.enqueue(collect_insert_dbstats)
    # needed for RQ Scheduler
    from rq_scheduler import Scheduler
    # Get a scheduler for the "default" queue
    scheduler = Scheduler(connection=Redis())

    from datetime import timedelta
    scheduler.enqueue_in(timedelta(minutes=3), collect_insert_dbstats)
    # from datetime import datetime
    # scheduler.schedule(
    #     scheduled_time=datetime.utcnow(),
    #     func=collect_insert_dbstats,
    #     interval=180,
    #     repeat=None
    # )
    # to cancel a scheduled job
    # scheduler.cancel('492a3e0f-6ac4-4f8a-8840-e9d4e9f983c7')
    return jsonify('task scheduled - stats will be collected and inserted')


@api.route('/flagme', methods=['POST'])
@http_auth.login_required
def flagme():
    # get the content of the POSTed body
    json_data = request.get_json()
    if not json_data:
        return jsonify({'message': 'No input data provided'}), 400
    # collect the POSTed data
    title = json_data['title']
    description = json_data['description']
    label = "WebGUI flag"
    # needed for RQ Scheduler
    from rq_scheduler import Scheduler
    # Get a scheduler for the "default" queue
    scheduler = Scheduler(connection=Redis())

    from datetime import timedelta
    scheduler.enqueue_in(timedelta(minutes=3), flag_to_gitlab, title, description, label)
    return jsonify('task scheduled - the issue will be added for review')


@api.route('/ping', methods=['GET'])
def ping():
    """ Just a simple route for checking the API service status
    Can be used within our monitoring machine, or to get notifications
    within Slack, ...
    """
    return jsonify(data="Pong!")


@api.route('/newtoken/', methods=['GET'])
def get_token():
    """ A simple route to generate a new token, for development or if somebody needs a token
    with a validity period different from the standard one for anonymous users: 24 hours
    """
    # let's generate a new token, with a random/sample user
    user_for_token = "sampleuser"
    email_for_token = "email@example.com"
    # token with 30 days validity period: 2592000
    # or for a token valid 5 years: 157680000
    expires_in = 2592000
    new_token = generate_auth_token(user_for_token, expires_in)
    # to have the token printed to the terminal, as well - for development
    # print(new_token)
    # let's verify the generated token is valid and get the username from it to prove it
    username = verify_auth_token(new_token)
    # print(username)
    return jsonify({"status": "Token Created", "user": {"username": username, "email": email_for_token},
                    "token": {"data": new_token.decode('ascii'), "expires_in": expires_in,
                              "user_from_token": user_for_token}})


######################
# SAML specific routes
######################

@api.route('/saml/', methods=['GET', 'POST'])
def index():
    req = prepare_flask_request(request)
    auth = init_saml_auth(req)
    errors = []
    not_auth_warn = False
    success_slo = False
    attributes = False
    paint_logout = False

    # SSO action (SP-SSO initiated). Will send an AuthNRequest to the IdP
    # initiate the login for API and redirect to user details if login is successful
    if 'sso' in request.args:
        return_to = url_for('api.user_attrs')
        return redirect(auth.login(return_to))
        # return redirect(auth.login())
    # another SSO init action with a custom RelayState URL
    # initiate the login for GUI and redirect to user back to GUI if successful
    elif 'sso2' in request.args:
        # based on machine's hostname use different base URL for redirection
        # TODO switch to use environment variables: PRODUCTION, TESTING, DEVELOPMENT
        if hostname == 'enc-production-nomad':
            return_to = 'https://encyclopedia.nomad-coe.eu/gui/'
        elif hostname == 'enc-production-hv-nomad':
            return_to = 'https://encyclopedia.nomad-coe.eu/gui/'
        else:
            return_to = '%sgui/' % request.host_url
        # return_to = '%sgui/' % request.host_url
        return redirect(auth.login(return_to))
        # return_to = url_for('api.user_attrs')
        # return redirect(auth.login(return_to))
    elif 'slo' in request.args:
        name_id = None
        session_index = None
        if 'samlNameId' in session:
            name_id = session['samlNameId']
        if 'samlSessionIndex' in session:
            session_index = session['samlSessionIndex']

        return redirect(auth.logout(name_id=name_id, session_index=session_index))
    # Assertion Consumer Service
    # handles the SAML response that the IdP forwards to the SP through the user's client
    elif 'acs' in request.args:
        # Process the Response of the IdP
        auth.process_response()
        # This method receives an array with the errors
        errors = auth.get_errors()
        not_auth_warn = not auth.is_authenticated()
        if len(errors) == 0:
            # Retrieves user data
            session['samlUserdata'] = auth.get_attributes()
            session['samlNameId'] = auth.get_nameid()
            session['samlSessionIndex'] = auth.get_session_index()
            self_url = OneLogin_Saml2_Utils.get_self_url(req)
            # user is authenticated, let's generate an API access token
            # first get the username received from IdP server
            username = session['samlUserdata']["urn:oid:0.9.2342.19200300.100.1.1"][0]
            # get the user email received from IdP server
            user_email = session['samlUserdata']["urn:oid:0.9.2342.19200300.100.1.3"][0]
            # having the username, generate the token having the username encoded in it
            # save the token to the session to access it in a different view: /saml/user
            # token expiration time set to 24h
            session['token_expires_in'] = 86400
            session['token'] = generate_auth_token(username, session['token_expires_in'])
            token = session['token']
            # Redirect if there is a RelayState
            if 'RelayState' in request.form and self_url != request.form['RelayState']:
                import json
                redirect_to_relaystate = redirect(auth.redirect_to(request.form['RelayState']))
                response = make_response(redirect_to_relaystate)
                cookie_data = {"status": "Authenticated",
                               "user": {"username": username, "email": user_email},
                               "token": {"data": token.decode('ascii'),
                                         "expires_in": session['token_expires_in']}}
                # production and staging machines have different domain names
                # and the cookie should be set on the corresponding domain name
                if hostname == 'enc-production-nomad' or hostname == 'enc-production-hv-nomad':
                    response.set_cookie('user_info', json.dumps(cookie_data), domain='.nomad-coe.eu')
                else:
                    response.set_cookie('user_info', json.dumps(cookie_data), domain='.mpg.de')
                return response
            # # user is authenticated, let's generate an API access token
            # # first get the username received from IdP server
            # username = session['samlUserdata']["urn:oid:0.9.2342.19200300.100.1.1"][0]
            # # having the username, generate the token having the username encoded in it
            # # save the token to the session to access it in a different view: /saml/user
            # # token expiration time set to 24h
            # session['token_expires_in'] = 86400
            # session['token'] = generate_auth_token(username, session['token_expires_in'])
            # return redirect(url_for('api.user_attrs'))
    # Single Logout Service (not used right now. /saml/logout endpoint is used)
    elif 'sls' in request.args:
        dscb = lambda: session.clear()
        url = auth.process_slo(delete_session_cb=dscb)
        errors = auth.get_errors()
        if len(errors) == 0:
            if url is not None:
                return redirect(url)
            else:
                success_slo = True
        # delete_session_callback = lambda: request.session.flush()
        # url = auth.process_slo(delete_session_cb=delete_session_callback)
        # errors = auth.get_errors()
        # if len(errors) == 0:
        #     if url is not None:
        #         print(url)
        #         return redirect(url)
        #     else:
        #         print("Sucessfully Logged out")
        # else:
        #     print("Error when processing SLO: %s" % (', '.join(errors)))

    if 'samlUserdata' in session:
        paint_logout = True
        if len(session['samlUserdata']) > 0:
            attributes = session['samlUserdata'].items()

    # return render_template(
    #     'index.html',
    #     errors=errors,
    #     not_auth_warn=not_auth_warn,
    #     success_slo=success_slo,
    #     attributes=attributes,
    #     paint_logout=paint_logout
    # )
    if 'samlUserdata' in session:
        # Added 'api' string as a quick fix for redirection to /saml/user/ after the user is authenticated.
        # Before the fix the url was domain.com/v1.0/saml/user/ which doesn't exists (because should be
        # domain.com/api/v1.0/saml/user/) and nginx redirected to the /gui and not showing the user's details, token ...
        return redirect('api' + url_for('api.user_attrs'))
    else:
        return redirect(auth.login())


@api.route('/saml/user/')
def user_attrs():
    paint_logout = False
    attributes = False
    if 'samlUserdata' in session:
        paint_logout = True
        if len(session['samlUserdata']) > 0:
            # attributes = session['samlUserdata'].items()
            attributes = session['samlUserdata']
            # let's take only the attributes we need. Each attribute is an array
            # by default in IdP, so, we take the first item of the array
            user_email = attributes["urn:oid:0.9.2342.19200300.100.1.3"][0]
            user_login = attributes["urn:oid:0.9.2342.19200300.100.1.1"][0]
            token = session['token']
            expires_in = session['token_expires_in']
            # next line is added for debug only, to be removed soon
            user_from_token = verify_auth_token(token)
            # if needed to show also all available user attributes, uncomment next line
            # return jsonify(user_data=attributes, username=user_login, email=user_email)
            # return jsonify(username=user_login, email=user_email,
            #                token=token.decode('ascii'), expires_in=expires_in,
            #                user_from_token=user_from_token)
            return jsonify({"status": "Authenticated",
                            "user": {"username": user_login, "email": user_email},
                            "token": {"data": token.decode('ascii'),
                                      "expires_in": expires_in}})
    else:
        return jsonify({"status": "Unauthenticated"})
    # return render_template('attrs.html', paint_logout=paint_logout,
    #                        attributes=attributes)


@api.route('/saml/logout/')
def logout():
    # the single logout is not working right now
    # just clear the session - not a final solution
    if 'samlUserdata' in session:
        session.clear()
        # idp_logout_url = 'https://nomad-login.csc.fi/idp/profile/Logout'
        idp_logout_url = 'https://idp.nomad-coe.eu/idp/profile/Logout'
        return redirect(idp_logout_url)

    return jsonify({"status": "Unauthenticated"})
    # if 'samlUserdata' in session:
    #     paint_logout = True
    #     success_slo = True
    #     if len(session['samlUserdata']) > 0:
    #         attributes = session['samlUserdata'].items()
    #
    # return render_template(
    #     'index.html',
    #     # errors=errors,
    #     # not_auth_warn=not_auth_warn,
    #     success_slo=success_slo,
    #     attributes=attributes,
    #     paint_logout=paint_logout
    # )


@api.route('/saml/metadata/')
def metadata():
    req = prepare_flask_request(request)
    auth = init_saml_auth(req)
    settings = auth.get_settings()
    metadata = settings.get_sp_metadata()
    errors = settings.validate_metadata(metadata)

    if len(errors) == 0:
        resp = make_response(metadata, 200)
        resp.headers['Content-Type'] = 'text/xml'
    else:
        resp = make_response(', '.join(errors), 500)
    return resp


#############################
# END OF SAML specific routes
#############################
