# Copyright 2016-2018 Ioan Vancea, Markus Scheidgen
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from flask import url_for, request


def paginate(collection, func_name=None, max_per_page=25):
    """Generate a paginated response for a resource collection.

    Routes that use this decorator must return a SQLAlchemy query as a
    response.

    The output of this decorator is a Python dictionary with the paginated
    results. The application must ensure that this result is converted to a
    response object, either by chaining another decorator or by using a
    custom response object that accepts dictionaries.
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            # invoke the wrapped function
            if request.method == 'POST':
                json_data = request.get_json()
                # page = int(json_data['search_by']['page'])
                # Return the value for a key if key is in the dictionary, else
                # default = 1. If default is not given, it defaults to None
                page = int(json_data['search_by'].get('page', "1"))
                # per_page = int(json_data['search_by']['per_page'])
                per_page = int(json_data['search_by'].get('per_page', max_per_page))
                if json_data['search_by'].get('pagination') == 'off':
                    page = 1
                    per_page = 10000
                # pagination = json_data['search_by']['pagination']
            query = f(*args, **kwargs)

            if request.method == 'GET':
                # obtain pagination arguments from the URL's query string
                page = request.args.get('page', 1, type=int)
                if request.args.get('pagination') == 'off':
                    page = 1
                    per_page = 10000
                else:
                    per_page = min(request.args.get('per_page', max_per_page,
                                                    type=int), max_per_page)

            # run the query with Flask-SQLAlchemy's pagination
            p = query.paginate(page, per_page)

            # build the pagination metadata to include in the response
            pages = {'page': page, 'per_page': per_page,
                     'total': p.total, 'pages': p.pages}

            if request.method == 'GET':
                if p.has_prev:
                    pages['prev_url'] = url_for(request.endpoint, page=p.prev_num,
                                                per_page=per_page,
                                                _external=True, **kwargs)
                else:
                    pages['prev_url'] = None
                if p.has_next:
                    pages['next_url'] = url_for(request.endpoint, page=p.next_num,
                                                per_page=per_page,
                                                _external=True, **kwargs)
                else:
                    pages['next_url'] = None

                pages['first_url'] = url_for(request.endpoint, page=1,
                                             per_page=per_page,
                                             _external=True, **kwargs)
                pages['last_url'] = url_for(request.endpoint, page=p.pages,
                                            per_page=per_page,
                                            _external=True, **kwargs)

            # generate the paginated collection as a dictionary
            # results = [(m.export_data(), calcs) for m, calcs in p.items]
            results = func_name.dump(p.items).data
            # results = m_calc_schema.dump(p.items).data
            # results = [item.export_data() for item in p.items]
            # results = [item.id for item in p.items]

            # "pages" should be "null" if no items for one complete page
            if p.total <= per_page:
                pages = None
            # return a dictionary as a response
            return {'results': results, 'pages': pages, 'total_results': p.total}
        return wrapped
    return decorator
