# Copyright 2016-2018 Ioan Vancea, Markus Scheidgen
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from flask import json as json_flask
from . import db, ma
from sqlalchemy import func
from sqlalchemy.dialects.postgresql import JSON
# from .utils import sum_form
from app.config import schema

############################
# SQLAlchemy Model
############################


class Material(db.Model):
    __tablename__ = 'material'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    material_name = db.Column(db.Text, index=True)
    periodicity = db.Column(db.Text)
    formula = db.Column(db.Text, nullable=False, index=True)
    formula_reduced = db.Column(db.Text, nullable=False, index=True)
    space_group_number = db.Column(db.Integer, index=True)
    space_group_international_short_symbol = db.Column(db.Text)
    point_group = db.Column(db.Text, index=True)
    bravais_lattice = db.Column(db.Enum('aP', 'mP', 'mS', 'oP',
                                        'oS', 'oI', 'oF', 'tP',
                                        'tI', 'cP', 'cI', 'cF',
                                        'hP', 'hR', 'rP', name='bravais_t'), nullable=False, index=True)
    crystal_system = db.Column(db.Enum('triclinic', 'monoclinic', 'orthorhombic',
                                       'tetragonal', 'trigonal', 'hexagonal',
                                       'cubic', name='crystal_t'), nullable=False, index=True)
    structure_type = db.Column(db.Text)
    structure_prototype = db.Column(db.Text)
    strukturbericht_designation = db.Column(db.Text)
    system_type = db.Column(db.Enum('unknown', 'atom', 'bulk', 'surface', 'molecule',
                                    '1D', '2D', 'polymer', name='system_t'), nullable=False, index=True)
    cell_angles_string = db.Column(db.Text, nullable=False)
    has_free_wyckoff_parameters = db.Column(db.Boolean, nullable=False)
    material_hash = db.Column(db.Text, nullable=False)

    element = db.relationship('Element', backref='material', lazy='dynamic')
    cell = db.relationship('Cell', backref='material', lazy='dynamic')
    material_classification = db.relationship('MaterialClassification', backref='material', lazy='dynamic')
    calc = db.relationship('Calc', backref='material', lazy='dynamic')

    def export_data(self):
        return {
            'id': self.id,
            'material_name': self.material_name,
            'periodicity': self.periodicity,
            'formula': self.formula,
            'formula_reduced': self.formula_reduced,
            'space_group_number': self.space_group_number,
            'point_group': self.point_group,
            'bravais_lattice': self.bravais_lattice,
            'crystal_system': self.crystal_system,
            'structure_type': self.structure_type,
            'structure_prototype': self.structure_prototype,
            'strukturbericht_designation': self.strukturbericht_designation,
            'system_type': self.system_type,
            'cell_angles_string': self.cell_angles_string,
            'has_free_wyckoff_parameters': self.has_free_wyckoff_parameters,
            'material_hash': self.material_hash
        }


class Element(db.Model):
    __tablename__ = 'element'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    label = db.Column(db.Integer, index=True)
    position = db.Column(db.String(256), nullable=False)
    wyckoff = db.Column(db.Text, nullable=False)
    material_id = db.Column(db.ForeignKey(
        '%s.material.id' % schema, ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    def export_data(self):
        return {
            'id': self.id,
            'label': self.label,
            'position': self.position,
            'wyckoff': self.wyckoff,
            'material_id': self.material_id
        }


class Cell(db.Model):
    __tablename__ = 'cell'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    is_primitive = db.Column(db.Boolean, nullable=False)
    a = db.Column(db.String(256), nullable=False)
    b = db.Column(db.String(256), nullable=False)
    c = db.Column(db.String(256), nullable=False)
    material_id = db.Column(db.ForeignKey(
        '%s.material.id' % schema, ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    def export_data(self):
        return {
            'id': self.id,
            'a': self.a,
            'b': self.b,
            'c': self.c,
            'is_primitive': self.is_primitive,
            'material_id': self.material_id
        }


class MaterialClassification(db.Model):
    __tablename__ = 'material_classification'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Text, nullable=False)
    class_type = db.Column(db.Enum('material_class_springer',
                                   'compound_class_springer', name='material_class_t'), nullable=False)
    material_id = db.Column(db.ForeignKey(
        '%s.material.id' % schema, ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    def export_data(self):
        return {
            'id': self.id,
            'name': self.name,
            'class_type': self.class_type,
            'material_id': self.material_id
        }


class Calc(db.Model):
    __tablename__ = 'calc'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    run_type = db.Column(db.Enum('single point',
                                 'molecular dynamics',
                                 'geometry optimization',
                                 'phonon calculation', name='run_type_t'), nullable=False, index=True)
    code_name = db.Column(db.Text, nullable=False, index=True)
    code_version = db.Column(db.Text)
    basis_set_type = db.Column(db.Enum('numeric AOs', 'gaussians', '(L)APW+lo',
                                       'plane waves', name='basis_set_t'), nullable=False, index=True)
    basis_set_short_name = db.Column(db.Text)
    core_electron_treatment = db.Column(db.Enum('none', 'full all electron', 'all electron frozen core',
                                                'pseudopotential', name='core_el_t'), nullable=False, index=True)
    pseudopotential_type = db.Column(db.Text)
    functional_type = db.Column(db.Enum('LDA', 'GGA', 'meta-GGA',
                                        'hybrid-GGA', 'hybrid-meta-GGA',
                                        'HF', 'vdW-DF', 'GW', name='funct_t'), nullable=False, index=True)
    functional_long_name = db.Column(db.Text, nullable=False)
    scf_threshold = db.Column(db.Float, nullable=False)
    band_gap = db.Column(db.Float, index=True)
    band_gap_direct = db.Column(db.Boolean, index=True)
    cell_volume = db.Column(db.Float(53), nullable=False)
    atomic_density = db.Column(db.Float(53), nullable=False)
    mass_density = db.Column(db.Float(53), nullable=False)
    k_point_grid_description = db.Column(db.Text)
    lattice_parameters = db.Column(db.String(256), nullable=False)
    pressure = db.Column(db.Float(53), nullable=False)
    smearing = db.Column(db.String(256), nullable=False)
    fermi_surface = db.Column(db.Text)
    calculation_pid = db.Column(db.Text)
    repository_download_uri = db.Column(db.Text)
    repository_upload_comment = db.Column(db.Text, nullable=False)
    has_dos = db.Column(db.Boolean, nullable=False)
    has_band_structure = db.Column(db.Boolean, nullable=False)
    has_fermi_surface = db.Column(db.Boolean, nullable=False)
    is_curtarolo_path = db.Column(db.Boolean, nullable=False)
    brillouin_zone_json = db.Column(db.Text, nullable=False)
    has_thermal_properties = db.Column(db.Boolean, nullable=False)
    gw_type = db.Column(db.Enum('', 'G0W0', 'ev-scGW',
                                'qp-scGW', 'scGW0', 'scG0W',
                                'scGW', name='gw_type_t'), nullable=False, index=True)
    elastic_calc_data_json = db.Column(db.Text)
    wyckoff_groups_json = db.Column(db.Text, nullable=False)
    settings_basis_set_json = db.Column(db.Text, nullable=False)
    method_hash = db.Column(db.Text, nullable=False)
    group_eos_hash = db.Column(db.Text, nullable=False)
    group_parametervariation_hash = db.Column(db.Text, nullable=False)
    mainfile_uri = db.Column(db.Text, nullable=False)
    src_filepath = db.Column(db.Text, nullable=False)

    material_id = db.Column(db.ForeignKey(
        '%s.material.id' % schema, ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    energy = db.relationship('EnergyDecomposition', backref='calc')
    contributor = db.relationship('Contributor', backref='calc')
    calculation_reference = db.relationship('CalculationReference', backref='calc')
    diagram_data = db.relationship('DiagramData', backref='calc')

    def export_data(self):
        return {
            'id': self.id,
            'run_type': self.run_type,
            'code_name': self.code_name,
            'code_version': self.code_version,
            'basis_set_type': self.basis_set_type,
            'basis_set_short_name': self.basis_set_short_name,
            'core_electron_treatment': self.core_electron_treatment,
            'pseudopotential_type': self.pseudopotential_type,
            'functional_type': self.functional_type,
            'functional_long_name': self.functional_long_name,
            'scf_threshold': self.scf_threshold,
            'band_gap': self.band_gap,
            'band_gap_direct': self.band_gap_direct,
            'cell_volume': self.cell_volume,
            'atomic_density': self.atomic_density,
            'mass_density': self.mass_density,
            'k_point_grid_description': self.k_point_grid_description,
            'lattice_parameters': self.lattice_parameters,
            'pressure': self.pressure,
            'smearing': self.smearing,
            'fermi_surface': self.fermi_surface,
            'calculation_pid': self.calculation_pid,
            'repository_download_uri': self.repository_download_uri,
            'repository_upload_comment': self.repository_upload_comment,
            'has_dos': self.has_dos,
            'has_band_structure': self.has_band_structure,
            'has_fermi_surface': self.has_fermi_surface,
            'is_curtarolo_path': self.is_curtarolo_path,
            'brillouin_zone_json': self.brillouin_zone_json,
            'has_thermal_properties': self.has_thermal_properties,
            'gw_type': self.gw_type,
            'elastic_calc_data_json': self.elastic_calc_data_json,
            'wyckoff_groups_json': self.wyckoff_groups_json,
            'settings_basis_set_json': self.settings_basis_set_json,
            'method_hash': self.method_hash,
            'group_eos_hash': self.group_eos_hash,
            'group_parametervariation_hash': self.group_parametervariation_hash,
            'mainfile_uri': self.mainfile_uri,
            'src_filepath': self.src_filepath,
            'material_id': self.material_id
        }


class EnergyDecomposition(db.Model):
    __tablename__ = 'energy'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    e_val = db.Column(db.Float(53), nullable=False)
    e_kind = db.Column(db.Text)
    calc_id = db.Column(db.ForeignKey('%s.calc.id' % schema,
                                      ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    def export_data(self):
        return {
            'id': self.id,
            'e_val': self.e_val,
            'e_kind': self.e_kind,
            'calc_id': self.calc_id
        }


class Contributor(db.Model):
    __tablename__ = 'contributor'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    first_name = db.Column(db.Text, nullable=False, index=True)
    last_name = db.Column(db.Text, nullable=False, index=True)
    contributor_type = db.Column(db.Enum(
        'uploader', 'coauthor', name='contributor_t'), nullable=False, index=True)
    calc_id = db.Column(db.ForeignKey('%s.calc.id' % schema,
                                      ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    def export_data(self):
        return {
            'id': self.id,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'contributor_type': self.contributor_type,
            'calc_id': self.calc_id
        }


class CalculationReference(db.Model):
    __tablename__ = 'calculation_reference'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    ref_target = db.Column(db.Text, nullable=False)
    ref_type = db.Column(db.Text, nullable=False)
    calc_id = db.Column(db.ForeignKey('%s.calc.id' % schema,
                                      ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    def export_data(self):
        return {
            'id': self.id,
            'ref_target': self.ref_target,
            'ref_type': self.ref_type,
            'calc_id': self.calc_id
        }


class DiagramData(db.Model):
    __tablename__ = 'diagram_data'
    __table_args__ = {'schema': schema}

    id = db.Column(db.Integer, primary_key=True)
    data_content = db.Column(db.Text, nullable=False)
    data_type = db.Column(db.Enum('dos', 'band_structure',
                                  'phonon_dos', 'phonon_dispersion',
                                  'specific_heat_cv', 'helmholtz_free_energy',
                                  'qha_bulk_modulus',
                                  'qha_helmholtz_free_energy',
                                  'qha_mass_density',
                                  'qha_specific_heat_cv',
                                  'qha_thermal_expansion',
                                  'gw_spectral_function', name='diagram_data_t'), nullable=False, index=True)
    calc_id = db.Column(db.ForeignKey('%s.calc.id' % schema,
                                      ondelete='CASCADE', onupdate='CASCADE'), nullable=False)

    def export_data(self):
        return {
            'id': self.id,
            'data_content': self.data_content,
            'data_type': self.data_type,
            'calc_id': self.calc_id
        }


############################
# Marshmallow Models
############################

class MaterialSchema(ma.ModelSchema):

    class Meta:
        model = Material
        # to expose only certain fields use the form below or
        # use the constructor as shown a few lines below
        # fields = ('id', 'material_name', 'formula', 'mat_hash')


class ElementSchema(ma.ModelSchema):

    class Meta:
        model = Element


class CellSchema(ma.ModelSchema):

    class Meta:
        model = Cell


class EnergyDecompositionSchema(ma.ModelSchema):

    class Meta:
        model = EnergyDecomposition


class ContributorSchema(ma.ModelSchema):

    class Meta:
        model = Contributor


class DiagramDataSchema(ma.ModelSchema):

    class Meta:
        model = DiagramData

    data_content = ma.Method('get_data_content')

    def get_data_content(self, obj):
        return json_flask.loads(obj.data_content)


class CalcSchema(ma.ModelSchema):

    class Meta:
        model = Calc
    # band_structure_json = ma.Function(lambda obj: json_flask.loads(obj.band_structure_json))
    # quick fix if there is no data available for band_structure_json.
    # Right now is just an empty field in database.
    brillouin_zone_json = ma.Method('get_brillouin_zone_json')
    elastic_calc_data_json = ma.Method('get_elastic_calc_data_json')
    wyckoff_groups_json = ma.Method('get_wyckoff_groups_json')
    settings_basis_set_json = ma.Method('get_settings_basis_set_json')
    dos = ma.Method('get_dos')
    band_structure = ma.Method('get_band_structure')
    phonon_dos = ma.Method('get_phonon_dos')
    phonon_dispersion = ma.Method('get_phonon_dispersion')
    specific_heat_cv = ma.Method('get_specific_heat_cv')
    helmholtz_free_energy = ma.Method('get_helmholtz_free_energy')
    qha_bulk_modulus = ma.Method('get_qha_bulk_modulus')
    qha_helmholtz_free_energy = ma.Method('get_qha_helmholtz_free_energy')
    qha_mass_density = ma.Method('get_qha_mass_density')
    qha_specific_heat_cv = ma.Method('get_qha_specific_heat_cv')
    qha_thermal_expansion = ma.Method('get_qha_thermal_expansion')
    gw_spectral_function = ma.Method('get_gw_spectral_function')
    # temporar fix to handle wrong default scf_threshold values inserted in DB
    scf_threshold = ma.Method('get_scf_threshold')
    energy = ma.Nested(EnergyDecompositionSchema, many=True)
    diagram_data = ma.Nested(DiagramDataSchema, many=True)

    # TODO: REFACTOR
    # here is some duplicate code. There's no reason to have a method for each
    # thing. Just to have a single method that takes which thing to get as an
    # argument, and returns the right one. Possibly using getattr if Marshmallow
    # forces me to use named attributes.
    def get_brillouin_zone_json(self, obj):
        if obj.brillouin_zone_json == "":
            return None
        return json_flask.loads(obj.brillouin_zone_json)

    # def get_elastic_calc_data_json(self, obj):
    #     if obj.elastic_calc_data_json == "":
    #         return None
    #     return json_flask.loads(obj.elastic_calc_data_json)

    def get_elastic_calc_data_json(self, obj):
        if obj.elastic_calc_data_json is None:
            return None
        return json_flask.loads(obj.elastic_calc_data_json)

    def get_wyckoff_groups_json(self, obj):
        if obj.wyckoff_groups_json == "":
            return None
        return json_flask.loads(obj.wyckoff_groups_json)

    def get_settings_basis_set_json(self, obj):
        if obj.settings_basis_set_json == "":
            return None
        return json_flask.loads(obj.settings_basis_set_json)

    def get_scf_threshold(self, obj):
        if obj.scf_threshold == 1:
            return None

    def get_dos(self, obj):
        # if obj.dos == "":
        #     return None
        return json_flask.loads(obj.dos)

    def get_band_structure(self, obj):
        return json_flask.loads(obj.band_structure)

    def get_phonon_dos(self, obj):
        return json_flask.loads(obj.phonon_dos)

    def get_phonon_dispersion(self, obj):
        return json_flask.loads(obj.phonon_dispersion)

    def get_specific_heat_cv(self, obj):
        return json_flask.loads(obj.specific_heat_cv)

    def get_helmholtz_free_energy(self, obj):
        return json_flask.loads(obj.helmholtz_free_energy)

    def get_qha_bulk_modulus(self, obj):
        return json_flask.loads(obj.qha_bulk_modulus)

    def get_qha_helmholtz_free_energy(self, obj):
        return json_flask.loads(obj.qha_helmholtz_free_energy)

    def get_qha_mass_density(self, obj):
        return json_flask.loads(obj.qha_mass_density)

    def get_qha_specific_heat_cv(self, obj):
        return json_flask.loads(obj.qha_specific_heat_cv)

    def get_qha_thermal_expansion(self, obj):
        return json_flask.loads(obj.qha_thermal_expansion)

    def get_gw_spectral_function(self, obj):
        return json_flask.loads(obj.gw_spectral_function)


class SearchMaterialSchema(ma.ModelSchema):
    id = ma.Int()
    formula = ma.String()
    formula_reduced = ma.String()
    structure_type = ma.String()
    nr_of_calculations_matching_criteria = ma.Int()
    space_group_number = ma.Int()
    space_group_international_short_symbol = ma.String()
    material_name = ma.String()
    system_type = ma.String()
    calculations_list_matching_criteria = ma.List(ma.Int())
    nr_of_calculations = ma.Method('get_total_nr_calcs')

    def get_total_nr_calcs(self, id):
        # result = Calc.query.filter_by(material_id = id[0]).count()
        result = Calc.query.with_entities(func.count(Calc.id)).filter_by(material_id=id[0]).scalar()
        return result


class SearchGroupSchema(ma.ModelSchema):
    method_hash = ma.String()
    material_hash = ma.String()
    group_eos_hash = ma.String()
    group_parametervariation_hash = ma.String()
    group_type = ma.Method('set_group_type')
    nr_of_calculations = ma.Int()
    calculations_list = ma.List(ma.Int())
    representative_calculation_id = ma.Int()
    energy_minimum = ma.Float()

    def set_group_type(self, obj):
        if hasattr(obj, 'group_eos_hash'):
            return "equation of state"
        elif hasattr(obj, 'group_parametervariation_hash'):
            return "convergence"
        else:
            return "method"


class ESSearchMaterialSchema(ma.ModelSchema):
    id = ma.Int(attribute="material_id")
    formula = ma.String()
    formula_reduced = ma.String()
    structure_type = ma.String()
    # two options to get a certain field to be dumped with a different name
    # calculations = ma.Int(dump_to='nr_of_calculations')
    nr_of_calculations = ma.Int(attribute="calculations")
    space_group_number = ma.Int()
    material_name = ma.String()
    system_type = ma.String()
    # show the next quantities in the search results just during the
    # development, to debug things
    functional_type = ma.List(ma.String())
    basis_set_type = ma.List(ma.String())
    atom_labels = ma.List(ma.String())
    crystal_system = ma.String()
    code_name = ma.List(ma.String())
    mass_density = ma.List(ma.String())
    band_gap = ma.List(ma.String())
    has_band_structure = ma.Boolean()
    has_dos = ma.Boolean()
    has_fermi_surface = ma.Boolean()
    has_termal_properties = ma.Boolean()


es_search_material_schema = ESSearchMaterialSchema(many=True)
search_material_schema = SearchMaterialSchema(many=True)
search_groups_schema = SearchGroupSchema(many=True)

material_schema = MaterialSchema(exclude=('calc', 'cell', 'element', 'material_classification'))
# for exposing specific fields
# material_schema = MaterialSchema(only=('id', 'formula'))
m_material_schema = MaterialSchema(many=True)
element_schema = ElementSchema()
# m_element_schema = ElementSchema(many=True, exclude=('position',))
m_element_schema = ElementSchema(many=True)
cell_schema = CellSchema()
m_cell_schema = CellSchema(many=True)
calc_schema = CalcSchema(exclude=('contributor', 'has_band_structure', 'has_thermal_properties',
                                  'calculation_reference', 'has_dos', 'has_fermi_surface',
                                  'is_curtarolo_path', 'src_filepath', 'repository_download_uri'))
m_calc_schema = CalcSchema(many=True,
                           exclude=('contributor', 'band_structure_json', 'calculation_reference',
                                    'dos_json', 'phonon_dispersion_json', 'phonon_dos_json', 'diagram_data',
                                    'heat_capacity_cv_json', 'qhp_results_json', 'src_filepath',
                                    'repository_download_uri', 'elastic_calc_data_json', 'settings_basis_set_json'))
energy_decomposition_schema = EnergyDecompositionSchema()
m_energy_decomposition_schema = EnergyDecompositionSchema(many=True)
contributor_schema = ContributorSchema()
m_contributor_schema = ContributorSchema(many=True)
diagram_data_schema = DiagramDataSchema()
m_diagram_data_schema = DiagramDataSchema(many=True)


# SQLAlchemy and Marshmallow models needed for data history stats table
class Stats(db.Model):
    __tablename__ = 'stats'
    __table_args__ = {'schema': 'data_history'}

    id = db.Column(db.Integer, primary_key=True)
    date = db.Column(db.Date, default=datetime.date.today(),
                     nullable=False, index=True)
    summary = db.Column(JSON, nullable=False)


class StatsSchema(ma.ModelSchema):

    class Meta:
        model = Stats
    date = ma.DateTime(format="%d-%m-%Y")


stats_schema = StatsSchema()
m_stats_schema = StatsSchema(many=True)
